@@ -20,30 +20,31 @@
 
 #include "module.h"
 
-#include "net-sendbuffer.h"
-#include "signals.h"
-#include "rawlog.h"
-#include "misc.h"
+#include <irssi/src/core/misc.h>
+#include <irssi/src/core/net-sendbuffer.h>
+#include <irssi/src/core/network.h>
+#include <irssi/src/core/rawlog.h>
+#include <irssi/src/core/signals.h>
 
-#include "channels.h"
-#include "queries.h"
+#include <irssi/src/core/channels.h>
+#include <irssi/src/core/queries.h>
 
-#include "irc-nicklist.h"
-#include "irc-queries.h"
-#include "irc-servers-setup.h"
-#include "irc-servers.h"
-#include "irc-cap.h"
-#include "sasl.h"
+#include <irssi/src/irc/core/irc-nicklist.h>
+#include <irssi/src/irc/core/irc-queries.h>
+#include <irssi/src/irc/core/irc-servers-setup.h>
+#include <irssi/src/irc/core/irc-servers.h>
+#include <irssi/src/irc/core/irc-cap.h>
+#include <irssi/src/irc/core/sasl.h>
 
-#include "channels-setup.h"
-#include "channel-rejoin.h"
-#include "servers-idle.h"
-#include "servers-reconnect.h"
-#include "servers-redirect.h"
-#include "modes.h"
+#include <irssi/src/core/channels-setup.h>
+#include <irssi/src/irc/core/channel-rejoin.h>
+#include <irssi/src/irc/core/servers-idle.h>
+#include <irssi/src/core/servers-reconnect.h>
+#include <irssi/src/irc/core/servers-redirect.h>
+#include <irssi/src/irc/core/modes.h>
 
-#include "settings.h"
-#include "recode.h"
+#include <irssi/src/core/settings.h>
+#include <irssi/src/core/recode.h>
 
 #define DEFAULT_MAX_KICKS 1
 #define DEFAULT_MAX_MODES 3
@@ -55,12 +56,6 @@
 #define DEFAULT_CMDS_MAX_AT_ONCE 5
 #define DEFAULT_MAX_QUERY_CHANS 1 /* more and more IRC networks are using stupid ircds.. */
 
-/*
- * 63 is the maximum hostname length defined by the protocol.  10 is a common
- * username limit on many networks.  1 is for the `@'.
- */
-#define MAX_USERHOST_LEN (63 + 10 + 1)
-
 void irc_servers_reconnect_init(void);
 void irc_servers_reconnect_deinit(void);
 
@@ -78,17 +73,26 @@ static int ischannel_func(SERVER_REC *server, const char *data)
 	IRC_SERVER_REC *irc_server = (IRC_SERVER_REC *) server;
 	char *chantypes, *statusmsg;
 
+	g_return_val_if_fail(data != NULL, FALSE);
+
+	/* empty string is no channel */
+	if (*data == '\0')
+		return FALSE;
+
 	chantypes = g_hash_table_lookup(irc_server->isupport, "chantypes");
 	if (chantypes == NULL)
 		chantypes = "#&!+"; /* normal, local, secure, modeless */
+
 	statusmsg = g_hash_table_lookup(irc_server->isupport, "statusmsg");
-	if (statusmsg == NULL)
-		statusmsg = "@+";
+	if (statusmsg == NULL && strchr(chantypes, '@') == NULL)
+		statusmsg = "@";
 
-	while (strchr(statusmsg, *data) != NULL)
-		data++;
+	if (statusmsg != NULL)
+		data += strspn(data, statusmsg);
 
-	return strchr(chantypes, *data) != NULL;
+	/* strchr(3) considers the trailing NUL as part of the string, make sure
+	 * we didn't advance too much. */
+	return *data != '\0' && strchr(chantypes, *data) != NULL;
 }
 
 static char **split_line(const SERVER_REC *server, const char *line,
@@ -108,11 +112,14 @@ static char **split_line(const SERVER_REC *server, const char *line,
 	 * the code much simpler.  It's worth it.
 	 */
 	len -= strlen(recoded_start) + strlen(recoded_end);
+	g_warn_if_fail(len > 0);
 	if (len <= 0) {
 		/* There is no room for anything. */
 		g_free(recoded_start);
 		g_free(recoded_end);
-		return NULL;
+		lines = g_new(char *, 1);
+		lines[0] = NULL;
+		return lines;
 	}
 
 	lines = recode_split(server, line, target, len, onspace);
@@ -197,67 +204,47 @@ static char **split_message(SERVER_REC *server, const char *target,
 
 	/* length calculation shamelessly stolen from splitlong_safe.pl */
 	return split_line(SERVER(server), msg, target,
-			  510 - strlen(":! PRIVMSG  :") -
+			  ircserver->max_message_len - strlen(":! PRIVMSG  :") -
 			  strlen(ircserver->nick) - MAX_USERHOST_LEN -
 			  strlen(target));
 }
 
-static void server_init(IRC_SERVER_REC *server)
+static void server_init_2(IRC_SERVER_REC *server)
 {
 	IRC_SERVER_CONNECT_REC *conn;
 	char *address, *ptr, *username, *cmd;
-	GTimeVal now;
 
 	g_return_if_fail(server != NULL);
 
 	conn = server->connrec;
 
-	if (conn->proxy != NULL && conn->proxy_password != NULL &&
-	    *conn->proxy_password != '\0') {
-		cmd = g_strdup_printf("PASS %s", conn->proxy_password);
-		irc_send_cmd_now(server, cmd);
-		g_free(cmd);
-	}
-
-	if (conn->proxy != NULL && conn->proxy_string != NULL) {
-		cmd = g_strdup_printf(conn->proxy_string, conn->address, conn->port);
-		irc_send_cmd_now(server, cmd);
-		g_free(cmd);
-	}
-
-	if (conn->sasl_mechanism != SASL_MECHANISM_NONE)
-		cap_toggle(server, "sasl", TRUE);
-
-	cap_toggle(server, "multi-prefix", TRUE);
-
-	irc_send_cmd_now(server, "CAP LS");
-
 	if (conn->password != NULL && *conn->password != '\0') {
-                /* send password */
+		/* send password */
 		cmd = g_strdup_printf("PASS %s", conn->password);
 		irc_send_cmd_now(server, cmd);
 		g_free(cmd);
 	}
 
-        /* send nick */
+	/* send nick */
 	cmd = g_strdup_printf("NICK %s", conn->nick);
 	irc_send_cmd_now(server, cmd);
 	g_free(cmd);
 
 	/* send user/realname */
 	address = server->connrec->address;
-        ptr = strrchr(address, ':');
+	ptr = strrchr(address, ':');
 	if (ptr != NULL) {
 		/* IPv6 address .. doesn't work here, use the string after
 		   the last : char */
-		address = ptr+1;
+		address = ptr + 1;
 		if (*address == '\0')
 			address = "x";
 	}
 
 	username = g_strdup(conn->username);
 	ptr = strchr(username, ' ');
-	if (ptr != NULL) *ptr = '\0';
+	if (ptr != NULL)
+		*ptr = '\0';
 
 	cmd = g_strdup_printf("USER %s %s %s :%s", username, username, address, conn->realname);
 	irc_send_cmd_now(server, cmd);
@@ -269,9 +256,47 @@ static void server_init(IRC_SERVER_REC *server)
 		irc_send_cmd_now(server, cmd);
 		g_free(cmd);
 	}
+}
 
-	server->isupport = g_hash_table_new((GHashFunc) g_istr_hash,
-					    (GCompareFunc) g_istr_equal);
+static void server_init_1(IRC_SERVER_REC *server)
+{
+	IRC_SERVER_CONNECT_REC *conn;
+	char *cmd;
+
+	g_return_if_fail(server != NULL);
+
+	conn = server->connrec;
+
+	if (conn->proxy != NULL && conn->proxy_password != NULL && *conn->proxy_password != '\0') {
+		cmd = g_strdup_printf("PASS %s", conn->proxy_password);
+		irc_send_cmd_now(server, cmd);
+		g_free(cmd);
+	}
+
+	if (conn->proxy != NULL && conn->proxy_string != NULL) {
+		cmd = g_strdup_printf(conn->proxy_string, conn->address, conn->port);
+		irc_send_cmd_now(server, cmd);
+		g_free(cmd);
+	}
+
+	if (conn->sasl_mechanism != SASL_MECHANISM_NONE) {
+		irc_cap_toggle(server, CAP_SASL, TRUE);
+	}
+
+	irc_cap_toggle(server, CAP_MULTI_PREFIX, TRUE);
+	irc_cap_toggle(server, CAP_EXTENDED_JOIN, TRUE);
+	irc_cap_toggle(server, CAP_SETNAME, TRUE);
+	irc_cap_toggle(server, CAP_INVITE_NOTIFY, TRUE);
+	irc_cap_toggle(server, CAP_AWAY_NOTIFY, TRUE);
+	irc_cap_toggle(server, CAP_CHGHOST, TRUE);
+	irc_cap_toggle(server, CAP_ACCOUNT_NOTIFY, TRUE);
+	irc_cap_toggle(server, CAP_SELF_MESSAGE, TRUE);
+	irc_cap_toggle(server, CAP_SERVER_TIME, TRUE);
+	if (!conn->use_tls && (conn->starttls || !conn->disallow_starttls)) {
+		irc_cap_toggle(server, CAP_STARTTLS, TRUE);
+	}
+
+	server->isupport = g_hash_table_new((GHashFunc) i_istr_hash, (GCompareFunc) i_istr_equal);
 
 	/* set the standards */
 	g_hash_table_insert(server->isupport, g_strdup("CHANMODES"), g_strdup("beI,k,l,imnpst"));
@@ -281,9 +306,122 @@ static void server_init(IRC_SERVER_REC *server)
 
 	/* prevent the queue from sending too early, we have a max cut off of 120 secs */
 	/* this will reset to 1 sec after we get the 001 event */
-	g_get_current_time(&now);
-	memcpy(&((IRC_SERVER_REC *)server)->wait_cmd, &now, sizeof(GTimeVal));
-	((IRC_SERVER_REC *)server)->wait_cmd.tv_sec += 120;
+	server->wait_cmd = g_get_real_time();
+	server->wait_cmd += 120 * G_USEC_PER_SEC;
+
+	if (!conn->no_cap) {
+		signal_emit("server waiting cap ls", 2, server, CAP_LS_VERSION);
+		irc_send_cmd_now(server, "CAP LS " CAP_LS_VERSION);
+		/* to detect non-CAP servers, send this bogus join */
+		/* the : will make INSPIRCD respond with 451 instead of 461, too */
+		irc_send_cmd_now(server, "JOIN :");
+	}
+	if (conn->starttls)
+		irc_server_send_starttls(server);
+	else if (conn->no_cap)
+		server_init_2(server);
+}
+
+static void init_ssl_loop(IRC_SERVER_REC *server, GIOChannel *handle)
+{
+	int error;
+	server->connrec->starttls = 1;
+
+	if (server->starttls_tag) {
+		g_source_remove(server->starttls_tag);
+		server->starttls_tag = 0;
+	}
+
+	error = irssi_ssl_handshake(handle);
+	if (error == -1) {
+		server->connection_lost = TRUE;
+		server_disconnect((SERVER_REC *) server);
+		return;
+	}
+	if (error & 1) { /* wait */
+		server->starttls_tag =
+		    i_input_add(handle, error == 1 ? I_INPUT_READ : I_INPUT_WRITE,
+		                (GInputFunction) init_ssl_loop, server);
+		return;
+	}
+	/* continue */
+	rawlog_redirect(server->rawlog, "Now talking encrypted");
+	signal_emit("server connection switched", 1, server);
+	if (!server->cap_supported) {
+		server_init_2(server);
+	} else {
+		signal_emit("server cap continue", 1, server);
+	}
+
+	if (settings_get_bool("starttls_sts")) {
+		IRC_SERVER_SETUP_REC *ssetup = IRC_SERVER_SETUP(server_setup_find(
+		    server->connrec->address, server->connrec->port, server->connrec->chatnet));
+		if (ssetup != NULL) {
+			ssetup->starttls = STARTTLS_ENABLED;
+			server_setup_add((SERVER_SETUP_REC *) ssetup);
+		}
+	}
+}
+
+#include <irssi/src/core/line-split.h>
+void irc_server_send_starttls(IRC_SERVER_REC *server)
+{
+	g_return_if_fail(server != NULL);
+
+	g_warning("[%s] Now attempting STARTTLS", server->tag);
+	irc_send_cmd_now(server, "STARTTLS");
+}
+
+static void event_starttls(IRC_SERVER_REC *server, const char *data)
+{
+	GIOChannel *ssl_handle;
+
+	g_return_if_fail(server != NULL);
+
+	if (!IS_IRC_SERVER(server))
+		return;
+
+	if (server->handle->readbuffer != NULL &&
+	    !line_split_is_empty(server->handle->readbuffer)) {
+		char *str;
+		line_split("", -1, &str, &server->handle->readbuffer);
+	}
+	ssl_handle = net_start_ssl((SERVER_REC *) server);
+	if (ssl_handle != NULL) {
+		g_source_remove(server->readtag);
+		server->readtag = -1;
+		server->handle->handle = ssl_handle;
+		init_ssl_loop(server, server->handle->handle);
+	} else {
+		g_warning("net_start_ssl failed");
+	}
+}
+
+static void event_registerfirst(IRC_SERVER_REC *server, const char *data)
+{
+	g_return_if_fail(server != NULL);
+
+	if (!IS_IRC_SERVER(server))
+		return;
+
+	if (server->connected)
+		return;
+
+	if (!server->cap_supported && !server->connrec->starttls)
+		server_init_2(server);
+}
+
+static void event_capend(IRC_SERVER_REC *server)
+{
+	g_return_if_fail(server != NULL);
+
+	if (!IS_IRC_SERVER(server))
+		return;
+
+	if (server->connected)
+		return;
+
+	server_init_2(server);
 }
 
 SERVER_REC *irc_server_init_connect(SERVER_CONNECT_REC *conn)
@@ -304,9 +442,11 @@ SERVER_REC *irc_server_init_connect(SERVER_CONNECT_REC *conn)
 
 	if (server->connrec->port <= 0) {
 		server->connrec->port =
-			server->connrec->use_ssl ? 6697 : 6667;
+			server->connrec->use_tls ? 6697 : 6667;
 	}
 
+	server->max_message_len = MAX_IRC_MESSAGE_LEN;
+
 	server->cmd_queue_speed = ircconn->cmd_queue_speed > 0 ?
 		ircconn->cmd_queue_speed : settings_get_time("cmd_queue_speed");
 	server->max_cmds_at_once = ircconn->max_cmds_at_once > 0 ?
@@ -322,7 +462,7 @@ SERVER_REC *irc_server_init_connect(SERVER_CONNECT_REC *conn)
 		ircconn->max_whois : DEFAULT_MAX_WHOIS;
 	server->max_msgs_in_cmd = ircconn->max_msgs > 0 ?
 		ircconn->max_msgs : DEFAULT_MAX_MSGS;
-	server->connrec->use_ssl = conn->use_ssl;
+	server->connrec->use_tls = conn->use_tls;
 
 	modes_server_init(server);
 
@@ -334,6 +474,27 @@ void irc_server_connect(SERVER_REC *server)
 {
 	g_return_if_fail(server != NULL);
 
+	if (server->connrec->connect_handle != NULL) {
+		/* an existing handle from upgrade */
+		IRC_SERVER_CONNECT_REC *conn;
+		int tls_disconnect;
+
+		conn = ((IRC_SERVER_REC *) server)->connrec;
+		tls_disconnect = conn->use_tls || conn->starttls;
+
+		if (tls_disconnect) {
+			/* we cannot use it, it is encrypted. force a reconnect */
+			g_io_channel_unref(conn->connect_handle);
+			conn->connect_handle = NULL;
+			server->session_reconnect = FALSE;
+			server_connect_ref((SERVER_CONNECT_REC *) conn);
+			server_disconnect(server);
+			server_connect((SERVER_CONNECT_REC *) conn);
+			server_connect_unref((SERVER_CONNECT_REC *) conn);
+			return;
+		}
+	}
+
 	if (!server_start_connect(server)) {
                 server_connect_unref(server->connrec);
 		g_free(server);
@@ -403,11 +564,10 @@ static void sig_connected(IRC_SERVER_REC *server)
 		(QUERY_REC *(*)(SERVER_REC *, const char *)) irc_query_find;
 	server->nick_comp_func = irc_nickcmp_rfc1459;
 
-	server->splits = g_hash_table_new((GHashFunc) g_istr_hash,
-					  (GCompareFunc) g_istr_equal);
+	server->splits = g_hash_table_new((GHashFunc) i_istr_hash, (GCompareFunc) i_istr_equal);
 
-        if (!server->session_reconnect)
-		server_init(server);
+	if (!server->session_reconnect)
+		server_init_1(server);
 }
 
 static void isupport_destroy_hash(void *key, void *value)
@@ -417,6 +577,17 @@ static void isupport_destroy_hash(void *key, void *value)
 }
 
 static void sig_disconnected(IRC_SERVER_REC *server)
+{
+	if (!IS_IRC_SERVER(server))
+		return;
+
+	if (server->starttls_tag) {
+		g_source_remove(server->starttls_tag);
+		server->starttls_tag = 0;
+	}
+}
+
+static void sig_destroyed(IRC_SERVER_REC *server)
 {
 	GSList *tmp;
 
@@ -431,15 +602,23 @@ static void sig_disconnected(IRC_SERVER_REC *server)
 	g_slist_free(server->cmdqueue);
 	server->cmdqueue = NULL;
 
-	gslist_free_full(server->cap_active, (GDestroyNotify) g_free);
+	i_slist_free_full(server->cap_active, (GDestroyNotify) g_free);
 	server->cap_active = NULL;
 
-	gslist_free_full(server->cap_supported, (GDestroyNotify) g_free);
-	server->cap_supported = NULL;
+	if (server->cap_supported) {
+		g_hash_table_destroy(server->cap_supported);
+		server->cap_supported = NULL;
+	}
 
-	gslist_free_full(server->cap_queue, (GDestroyNotify) g_free);
+	i_slist_free_full(server->cap_queue, (GDestroyNotify) g_free);
 	server->cap_queue = NULL;
 
+	/* was g_free_and_null, but can't use on a GString */
+	if (server->sasl_buffer != NULL) {
+		g_string_free(server->sasl_buffer, TRUE);
+		server->sasl_buffer = NULL;
+	}
+
 	/* these are dynamically allocated only if isupport was sent */
 	g_hash_table_foreach(server->isupport,
 			     (GHFunc) isupport_destroy_hash, server);
@@ -485,7 +664,7 @@ char **irc_server_split_action(IRC_SERVER_REC *server, const char *target,
 	g_return_val_if_fail(data != NULL, NULL);
 
 	return split_line(SERVER(server), data, target,
-			  510 - strlen(":! PRIVMSG  :\001ACTION \001") -
+			  server->max_message_len - strlen(":! PRIVMSG  :\001ACTION \001") -
 			  strlen(server->nick) - MAX_USERHOST_LEN -
 			  strlen(target));
 }
@@ -519,27 +698,61 @@ void irc_server_send_data(IRC_SERVER_REC *server, const char *data, int len)
 		return;
 	}
 
-	g_get_current_time(&server->last_cmd);
+	server->last_cmd = g_get_real_time();
 
 	/* A bit kludgy way to do the flood protection. In ircnet, there
 	   actually is 1sec / 100 bytes penalty, but we rather want to deal
 	   with the max. 1000 bytes input buffer problem. If we send more
 	   than that with the burst, we'll get excess flooded. */
 	if (len < 100 || server->cmd_queue_speed <= 10)
-		server->wait_cmd.tv_sec = 0;
+		server->wait_cmd = 0;
 	else {
-		memcpy(&server->wait_cmd, &server->last_cmd, sizeof(GTimeVal));
-		server->wait_cmd.tv_sec += 2 + len/100;
+		server->wait_cmd = server->last_cmd;
+		server->wait_cmd += (2 + len / 100) * G_USEC_PER_SEC;
 	}
 }
 
-static int server_cmd_timeout(IRC_SERVER_REC *server, GTimeVal *now)
+void irc_server_send_and_redirect(IRC_SERVER_REC *server, GString *str, REDIRECT_REC *redirect)
+{
+	int crlf;
+
+	g_return_if_fail(server != NULL);
+	g_return_if_fail(str != NULL);
+
+	if (str->len > 2 && str->str[str->len - 2] == '\r')
+		crlf = 2;
+	else if (str->len > 1 && str->str[str->len - 1] == '\n')
+		crlf = 1;
+	else
+		crlf = 0;
+
+	if (crlf)
+		g_string_truncate(str, str->len - crlf);
+
+	signal_emit("server outgoing modify", 3, server, str, crlf);
+	if (str->len) {
+		if (crlf == 2)
+			g_string_append(str, "\r\n");
+		else if (crlf == 1)
+			g_string_append(str, "\n");
+
+		irc_server_send_data(server, str->str, str->len);
+
+		/* add to rawlog without [CR+]LF */
+		if (crlf)
+			g_string_truncate(str, str->len - crlf);
+		rawlog_output(server->rawlog, str->str);
+		server_redirect_command(server, str->str, redirect);
+	}
+}
+
+static int server_cmd_timeout(IRC_SERVER_REC *server, gint64 now)
 {
 	REDIRECT_REC *redirect;
-        GSList *link;
+	GSList *link;
+	GString *str;
 	long usecs;
 	char *cmd;
-	int len;
 
 	if (!IS_IRC_SERVER(server))
 		return 0;
@@ -547,31 +760,25 @@ static int server_cmd_timeout(IRC_SERVER_REC *server, GTimeVal *now)
 	if (server->cmdcount == 0 && server->cmdqueue == NULL)
 		return 0;
 
-	if (g_timeval_cmp(now, &server->wait_cmd) == -1)
+	if (now < server->wait_cmd)
 		return 1;
 
-	usecs = get_timeval_diff(now, &server->last_cmd);
+	usecs = (now - server->last_cmd) / G_TIME_SPAN_MILLISECOND;
 	if (usecs < server->cmd_queue_speed)
 		return 1;
 
 	server->cmdcount--;
-	if (server->cmdqueue == NULL) return 1;
+	if (server->cmdqueue == NULL)
+		return 1;
 
-        /* get command */
+	/* get command */
 	cmd = server->cmdqueue->data;
-        redirect = server->cmdqueue->next->data;
+	redirect = server->cmdqueue->next->data;
 
 	/* send command */
-	len = strlen(cmd);
-	irc_server_send_data(server, cmd, len);
-
-	/* add to rawlog without [CR+]LF */
-        if (len > 2 && cmd[len-2] == '\r')
-		cmd[len-2] = '\0';
-        else if (cmd[len-1] == '\n')
-		cmd[len-1] = '\0';
-	rawlog_output(server->rawlog, cmd);
-	server_redirect_command(server, cmd, redirect);
+	str = g_string_new(cmd);
+	irc_server_send_and_redirect(server, str, redirect);
+	g_string_free(str, TRUE);
 
 	/* remove from queue */
 	server->cmdqueue = g_slist_remove(server->cmdqueue, cmd);
@@ -586,13 +793,13 @@ static int server_cmd_timeout(IRC_SERVER_REC *server, GTimeVal *now)
 /* check every now and then if there's data to be sent in command buffer */
 static int servers_cmd_timeout(void)
 {
-	GTimeVal now;
+	gint64 now;
 	GSList *tmp;
 	int keep = 0;
 
-	g_get_current_time(&now);
+	now = g_get_real_time();
 	for (tmp = servers; tmp != NULL; tmp = tmp->next) {
-		keep |= server_cmd_timeout(tmp->data, &now);
+		keep |= server_cmd_timeout(tmp->data, now);
 	}
 	if (keep)
 		return 1;
@@ -611,45 +818,30 @@ void irc_servers_start_cmd_timeout(void)
 
 /* Return a string of all channels (and keys, if any have them) in server,
    like "#a,#b,#c,#d x,b_chan_key,x,x" or just "#e,#f,#g" */
-char *irc_server_get_channels(IRC_SERVER_REC *server)
+char *irc_server_get_channels(IRC_SERVER_REC *server, int rejoin_channels_mode)
 {
 	GSList *tmp;
 	GString *chans, *keys;
 	char *ret;
 	int use_keys;
-	char *rejoin_channels_mode;
 
 	g_return_val_if_fail(server != NULL, FALSE);
 
-	rejoin_channels_mode = g_strdup(settings_get_str("rejoin_channels_on_reconnect"));
-
-	if (rejoin_channels_mode == NULL ||
-	    (g_ascii_strcasecmp(rejoin_channels_mode, "on") != 0 &&
-	     g_ascii_strcasecmp(rejoin_channels_mode, "off") != 0 &&
-	     g_ascii_strcasecmp(rejoin_channels_mode, "auto") != 0)) {
-		g_warning("Invalid value for 'rejoin_channels_on_reconnect', valid values are 'on', 'off', 'auto', using 'on' as default value.");
-		g_free(rejoin_channels_mode);
-		rejoin_channels_mode = g_strdup("on");
-	}
+	/* do we want to rejoin channels in the first place? */
+	if (rejoin_channels_mode == REJOIN_CHANNELS_MODE_OFF)
+		return g_strdup("");
 
 	chans = g_string_new(NULL);
 	keys = g_string_new(NULL);
 	use_keys = FALSE;
 
-	/* do we want to rejoin channels in the first place? */
-	if(g_ascii_strcasecmp(rejoin_channels_mode, "off") == 0) {
-		g_string_free(chans, TRUE);
-		g_string_free(keys, TRUE);
-		g_free(rejoin_channels_mode);
-		return g_strdup("");
-	}
-
 	/* get currently joined channels */
 	for (tmp = server->channels; tmp != NULL; tmp = tmp->next) {
 		CHANNEL_REC *channel = tmp->data;
 		CHANNEL_SETUP_REC *setup = channel_setup_find(channel->name, channel->server->connrec->chatnet);
-		if ((setup != NULL && setup->autojoin && g_ascii_strcasecmp(rejoin_channels_mode, "auto") == 0) ||
-		    g_ascii_strcasecmp(rejoin_channels_mode, "on") == 0) {
+		if ((setup != NULL && setup->autojoin &&
+		     rejoin_channels_mode == REJOIN_CHANNELS_MODE_AUTO) ||
+		    rejoin_channels_mode == REJOIN_CHANNELS_MODE_ON) {
 			g_string_append_printf(chans, "%s,", channel->name);
 			g_string_append_printf(keys, "%s,", channel->key == NULL ? "x" : channel->key);
 			if (channel->key != NULL)
@@ -662,8 +854,9 @@ char *irc_server_get_channels(IRC_SERVER_REC *server)
 		REJOIN_REC *rec = tmp->data;
 		CHANNEL_SETUP_REC *setup = channel_setup_find(rec->channel, server->tag);
 
-		if ((setup != NULL && setup->autojoin && g_ascii_strcasecmp(rejoin_channels_mode, "auto") == 0) ||
-		    g_ascii_strcasecmp(rejoin_channels_mode, "on") == 0) {
+		if ((setup != NULL && setup->autojoin &&
+		     rejoin_channels_mode == REJOIN_CHANNELS_MODE_AUTO) ||
+		    rejoin_channels_mode == REJOIN_CHANNELS_MODE_ON) {
 			g_string_append_printf(chans, "%s,", rec->channel);
 			g_string_append_printf(keys, "%s,", rec->key == NULL ? "x" :
 									rec->key);
@@ -681,7 +874,6 @@ char *irc_server_get_channels(IRC_SERVER_REC *server)
 	ret = chans->str;
 	g_string_free(chans, FALSE);
 	g_string_free(keys, TRUE);
-	g_free(rejoin_channels_mode);
 
 	return ret;
 }
@@ -689,7 +881,6 @@ char *irc_server_get_channels(IRC_SERVER_REC *server)
 static void event_connected(IRC_SERVER_REC *server, const char *data, const char *from)
 {
 	char *params, *nick;
-	GTimeVal now;
 
 	g_return_if_fail(server != NULL);
 
@@ -712,8 +903,7 @@ static void event_connected(IRC_SERVER_REC *server, const char *data, const char
 	server->real_connect_time = time(NULL);
 
 	/* let the queue send now that we are identified */
-	g_get_current_time(&now);
-	memcpy(&server->wait_cmd, &now, sizeof(GTimeVal));
+	server->wait_cmd = g_get_real_time();
 
 	if (server->connrec->usermode != NULL) {
 		/* Send the user mode, before the autosendcmd.
@@ -1027,7 +1217,8 @@ void irc_server_init_isupport(IRC_SERVER_REC *server)
 
 void irc_servers_init(void)
 {
-	settings_add_str("servers", "rejoin_channels_on_reconnect", "on");
+	settings_add_bool("servers", "starttls_sts", TRUE);
+	settings_add_choice("servers", "rejoin_channels_on_reconnect", 1, "off;on;auto");
 	settings_add_str("misc", "usermode", DEFAULT_USER_MODE);
 	settings_add_str("misc", "split_line_start", "");
 	settings_add_str("misc", "split_line_end", "");
@@ -1038,8 +1229,12 @@ void irc_servers_init(void)
 	cmd_tag = -1;
 
 	signal_add_first("server connected", (SIGNAL_FUNC) sig_connected);
-	signal_add_last("server disconnected", (SIGNAL_FUNC) sig_disconnected);
+	signal_add_first("server disconnected", (SIGNAL_FUNC) sig_disconnected);
+	signal_add_last("server destroyed", (SIGNAL_FUNC) sig_destroyed);
 	signal_add_last("server quit", (SIGNAL_FUNC) sig_server_quit);
+	signal_add("event 670", (SIGNAL_FUNC) event_starttls);
+	signal_add("event 451", (SIGNAL_FUNC) event_registerfirst);
+	signal_add("server cap end", (SIGNAL_FUNC) event_capend);
 	signal_add("event 001", (SIGNAL_FUNC) event_connected);
 	signal_add("event 004", (SIGNAL_FUNC) event_server_info);
 	signal_add("event 005", (SIGNAL_FUNC) event_isupport);
@@ -1066,7 +1261,11 @@ void irc_servers_deinit(void)
 
 	signal_remove("server connected", (SIGNAL_FUNC) sig_connected);
 	signal_remove("server disconnected", (SIGNAL_FUNC) sig_disconnected);
+	signal_remove("server destroyed", (SIGNAL_FUNC) sig_destroyed);
         signal_remove("server quit", (SIGNAL_FUNC) sig_server_quit);
+	signal_remove("event 670", (SIGNAL_FUNC) event_starttls);
+	signal_remove("event 451", (SIGNAL_FUNC) event_registerfirst);
+	signal_remove("server cap end", (SIGNAL_FUNC) event_capend);
 	signal_remove("event 001", (SIGNAL_FUNC) event_connected);
 	signal_remove("event 004", (SIGNAL_FUNC) event_server_info);
 	signal_remove("event 005", (SIGNAL_FUNC) event_isupport);
